#!/usr/bin/env python

import os, sys
from os.path import exists, join, dirname, isdir, isabs
import shutil
import subprocess

from rainbow import util

USAGE = """

NAME:

  rainbow-inject

Purpose:

  rainbow-inject is a security-wrapper for Sugar Activities. Its purpose is to
  launch them in a low-privilege environment so that, if they are exploited,
  they are able to do less harm than would be the case if they ran with the
  full authority of the XO's human operator.

  This isolation is accomplished by running each activity as a separate user
  from a set of users that is known, at system installation time, to have
  tightly restricted access to the file-system.

  The environment that rainbow-inject creates for each activity it starts is
  controlled by an activity-specific security-record created by the
  rainbow-install-bundle.

Environment Variables:

  SUGAR_BUNDLE_PATH  : MUST contain an absolute path to a bundle to launch.

  SUGAR_BUNDLE_PATH  : MUST contain the unique bundle-id of the bundle at
                       SUGAR_BUNDLE_PATH.

  RAINBOW_DIR        : Optional; defaults to /activities
                       This directory is used by rainbow-inject for all its
                       run-time data: currently, this means uid/gid allocation
                       data and for storing persistent activity-data storage

  RAINBOW_STRACE_LOG : Optional; defaults to Nothing.
                       If set, rainbow-inject will cause strace to write a
                       complete log to the specified path.

Arguments:

  Any arguments will be exec'ed verbatim after rainbow-inject properly
  configures its environment.

File Descriptors:

  Open file-descriptors will be passed unmodified to the activity being
  launched. However, for the foreseeable future, rainbow-inject _will_ print
  trace information to descriptor 1 and may print error messages to descriptor
  2 for debugging purposes.

Return Codes:

    1: An unhandled exception arose.

"""

def log_file(filelike):
    def log(msg, *args):
        if len(args):
            filelike.write(msg % args + '\n')
        else:
            filelike.write(msg + '\n')
    return log

def strace(log, log_path, argv, env):
    # cjb thinks we should really be using SystemTap here and refers us to
    # wiki:SystemTap and #83 <MS>
    log('applying strace')
    args = ['/usr/bin/strace', '-f', '-o', log_path]
    for k, v in env.iteritems():
        args.append('-E')
        args.append('%s=%s' % (k, v))
    args.extend(argv)

    return args, {}

def grab_home(log, spool, bundle_id, owner_uid, owner_gid):
    uid = util.grab_uid(join(spool, 'uid_pool'))
    gid = util.grab_gid(join(spool, 'gid_pool'), join(spool, 'bundle_id_to_gid'), bundle_id)
    home = join(spool, 'uid_to_home_dir', str(uid))
    comment = '%s.%d' % (bundle_id, uid)
    assert not exists(home)

    gid_cmd = ['/usr/sbin/groupadd', '-o', '-g', str(gid), str(gid)]
    log('adding group: %s', ' '.join(gid_cmd))
    try:
        subprocess.check_call(gid_cmd)
    except subprocess.CalledProcessError, e:
        if e.returncode is not 9: # we don't care if the group already exists
            raise e

    uid_cmd = ['/usr/sbin/useradd', '-m', '-u', str(uid), '-g', str(gid),
                                    '-c', comment, '-d', home, str(uid)]
    log('adding user: %s', ' '.join(uid_cmd))
    subprocess.check_call(uid_cmd)

    os.chown(home, uid, owner_gid)
    os.chmod(home, 0770)

    return (uid, gid, home)

def configure_home(log, spool, bundle_id, owner_uid, owner_gid, uid, gid, home):
    path = join(spool, 'uid_to_instance_dir', str(uid))
    util.make_dirs(path, uid, owner_gid, 0770)
    os.chown(path, uid, owner_gid)
    os.chmod(path, 0770)
    os.symlink(path, join(home, 'instance'))

    path = join(spool, 'gid_to_data_dir', str(gid))
    util.make_dirs(path, owner_uid, gid, 0770)
    os.chown(path, owner_uid, gid)
    os.chmod(path, 0770)
    os.symlink(path, join(home, 'data'))

    util.make_dirs(join(home, 'tmp'), uid, gid, 0700)
    util.mount('tmpfs', join(home, 'tmp'), 'tmpfs', 0, "size=1M,nr_inodes=1024")
    # home is 770 <uid>.<owner_gid> and, in the normal implementation, is in a
    # separate namespace.

    xauth = join(home, '.Xauthority')
    shutil.copy('/home/olpc/.Xauthority', xauth)
    os.chown(xauth, uid, gid)

    for frag in ['session.info', 'config']:
        path = join('.sugar/default/', frag)
        util.make_dirs(dirname(join(home, path)), uid, owner_gid, 0770)
        shutil.copy(join('/home/olpc', path),
                    join(home, path))
        os.chown(join(home, path), uid, owner_gid)
        os.chmod(join(home, path), 0770)

def launch(log, home, uid, gid, argv, env):
    env['USER'] = str(uid)
    env['HOME'] = home
    env['XAUTHORITY'] = join(home, '.Xauthority')
    env['SUGAR_ACTIVITY_ROOT'] = home
    env['TMPDIR'] = join(home, 'tmp')

    log('dropping privilege to (%d, %d)', uid, gid)
    os.setgid(gid)
    os.setuid(uid)

    log('chdir to %s' % home)
    os.chdir(home)

    log('umask(0)')
    os.umask(0)

    # Since we've dropped privilege, there's nothing to verify about strace_log
    strace_log = env.get('RAINBOW_STRACE_LOG')
    if strace_log:
        argv, env = strace(strace_log, argv, env)

    log('about to execve\nargv: %s\nenv: %s', argv, env)
    os.execvpe(argv[0], argv, env)

def verify_bundle_path(path):
    assert path and os.access(path, os.F_OK | os.R_OK) and isabs(path)
    return path

def verify_bundle_id(bundle_id):
    assert bundle_id and all(s.isalnum() for s in bundle_id.split('.'))
    return bundle_id

def verify_argv(argv):
    assert argv and len(argv) > 1
    return argv

def rwx_dir(path):
    return os.access(path, os.F_OK | os.R_OK | os.W_OK | os.X_OK) \
       and isdir(path)

def verify_spool(spool):
    assert spool
    for frag in ('uid_pool', 'gid_pool', 'bundle_id_to_gid',
                 'uid_to_instance_dir', 'gid_to_data_dir', 'uid_to_home_dir'):
        assert rwx_dir(join(spool, frag))

    return spool

def verify_home(home, uid, gid):
    saved_uid = os.getuid()
    saved_gid = os.getgid()
    os.setregid(-1, gid)
    os.setreuid(-1, uid)
    for frag in ('instance', 'data', 'tmp'):
        assert rwx_dir(join(home, frag))
    os.setreuid(-1, saved_uid)
    os.setregid(-1, saved_gid)

def verify_owner(owner_uid, owner_gid):
    assert (owner_uid, owner_gid) == (500, 500)
    return (owner_uid, owner_gid)

def run(log, env, argv, owner_uid, owner_gid, bundle_path, bundle_id, spool):
    # Note: exceptions are intended to bubble up to the caller and should
    # terminate execution.
    bundle_path = verify_bundle_path(bundle_path)
    bundle_id   = verify_bundle_id(bundle_id)
    argv        = verify_argv(argv)
    spool   = verify_spool(spool)
    owner_uid, owner_gid = verify_owner(owner_uid, owner_gid)

    (uid, gid, home) = grab_home(log, spool, bundle_id, owner_uid, owner_gid)
    configure_home(log, spool, bundle_id, owner_uid, owner_gid, uid, gid, home)

    verify_home(home, uid, gid)
    verify_home(home, owner_uid, owner_gid)

    launch(log, home, uid, gid, argv, env)

def main():
    run(log_file(sys.stdout),
        os.environ,
        sys.argv,
        int(os.environ.get('OWNER_UID', 500)),
        int(os.environ.get('OWNER_GID', 500)),
        os.environ.get('SUGAR_BUNDLE_PATH'),
        os.environ.get('SUGAR_BUNDLE_ID'),
        (os.environ.get('RAINBOW_DIR') or '/activities'))

if __name__ == '__main__':
    main()
